/*
Program:
  Semestralni prace z predmenu KIV/ZOS
	Problem vecericich filozofu

Autor:
	Antonin Neumann, A11B0439P
	neumann@students.zcu.cz
*/
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <semaphore.h>

#define MIN_PHIL 5 /* minimalni pocet filozofu */
#define MAX_PHIL 30 /* maximalni pocet filozofu */
#define MIN_EAT 2 /* minimalni pocet jezeni */
#define MAX_EAT 30 /* maximalni pocet jezeni */

/* semafory */
sem_t* vidlicky = NULL;
sem_t vidlicky_pristup;

/* globalni promenne */
int* jedl;
int pocet_jezeni;

/*pocet vlaken (filozofu)*/
int pocet_vlaken;

/* Funkce uspi program na nahodnou dobu v rozmeni 100 - 900 milisekund*/
void slofik(void){
	usleep(((rand() % 10) + 1) * 100000);
}

/* Funkce uspi program na zadany pocet sekund */
void sleep_sec(int second){
	int i;
	for(i = 0; i < second; i++){
		usleep(999999);
	}
}

/* funkcni prototypy */
void *filozof(void *argument);
void *deadlock(void *argument);
void *vyhladoveni(void *argument);
void *poradi(void *argument);
void help();



int main(int argc, char *argv[]){
	/* nastaveni poctu vlaken/filozofu */
	if(argc >= 2){
		if(strcmp(argv[1], "help") == 0){
			help();
			exit(0);
		}

		pocet_vlaken = atoi(argv[1]);
		if(pocet_vlaken < MIN_PHIL){
			printf("Filosofu musi byt alespon %d - prenastavuji...\n", MIN_PHIL);
			pocet_vlaken = MIN_PHIL;
		}
		if(pocet_vlaken > MAX_PHIL){
			printf("Filosofu smi byt maximalne %d - prenastavuji...\n", MAX_PHIL);
			pocet_vlaken = MAX_PHIL;
		}
	}
	else {
		printf("Nezadal jsi pocet filosofu, urcim je sam.\n");
		pocet_vlaken = MIN_PHIL;
	}
	printf("Pocet vlaken/filosofu je %d\n", pocet_vlaken);
  /***************************************************************/

	/* Nastaveni poctu jezeni */
	if(argc >= 3){
		pocet_jezeni = atoi(argv[2]);
		if(pocet_jezeni < MIN_EAT){
			printf("Pocet jezeni musi byt alespon %d - prenastavuji...\n", MIN_EAT);
			pocet_jezeni = MIN_EAT;
		}
		if(pocet_jezeni > MAX_EAT){
			printf("Pocet jezeni smi byt maximalne %d - prenastavuji...\n", MAX_EAT);
			pocet_jezeni = MAX_EAT;
		}
	}
	else {
		printf("Nezadal jsi pocet jezeni, nastavuji implicitni hodnotu\n");
		pocet_jezeni = MIN_EAT;
	}
	printf("Pocet pocet jezeni je nastaven na %d\n", pocet_jezeni);
	/***************************************************************/

	/* inicializace vlaken */
	pthread_t vlakna[pocet_vlaken];
	/* alokace mista v pameti pro vidlicky */
	vidlicky = malloc(pocet_vlaken*sizeof(sem_t));
	/* alokace mista v pameti pro pocet jezeni kazdeho filozofa */
	jedl = malloc(pocet_vlaken*sizeof(int));
	/* inicializace semaforu pro pristup k vidlickam obecne */
	sem_init(&vidlicky_pristup, 0, 1);

  int i;
	/* vyber typu symulace */
	if(argc >= 4 && strcmp(argv[3], "deadlock") == 0) {
    printf("Simulace deadlocku\n");
    sleep_sec(2);
		/* inicializace semaforu pro kazdou vidlicku a vlaken */
		for(i = 0; i < pocet_vlaken; i++)
		{
	  	sem_init(&(vidlicky[i]), 0, 1);
   		if(pthread_create(&(vlakna[i]), NULL, &deadlock, (void *) i) < 0)
			{
				perror("pthread_create");
				exit(1);
			}
		}
	}
	else if(argc >= 4 && (strcmp(argv[3], "starvation") == 0 || strcmp(argv[3], "vyhladoveni") == 0)){
    printf("Simulace vyhladoveni\n");
    sleep_sec(2);
		/* inicializace semaforu pro kazdou vidlicku a vlaken */
		for(i = 0; i < pocet_vlaken; i++)
		{
	  	sem_init(&(vidlicky[i]), 0, 1);
   		if(pthread_create(&(vlakna[i]), NULL, &vyhladoveni, (void *) i) < 0)
			{
				perror("pthread_create");
				exit(1);
			}
		}
	}
  else if(argc >= 4 && (strcmp(argv[3], "order") == 0 || strcmp(argv[3], "poradi") == 0)){
    printf("Simulace podle poradi\n");
		sleep_sec(2);
		/* inicializace semaforu pro kazdou vidlicku a vlaken */
		for(i = 0; i < pocet_vlaken; i++)
		{
	  	sem_init(&(vidlicky[i]), 0, 1);
   		if(pthread_create(&(vlakna[i]), NULL, &poradi, (void *) i) < 0)
			{
				perror("pthread_create");
				exit(1);
			}
		}
	}
	else {
    printf("Normalni simulace\n");
    sleep_sec(2);
		/* inicializace semaforu pro kazdou vidlicku a vlaken */
		for(i = 0; i < pocet_vlaken; i++)
		{
	  	sem_init(&(vidlicky[i]), 0, 1);
   		if(pthread_create(&(vlakna[i]), NULL, &filozof, (void *) i) < 0)
			{
				perror("pthread_create");
				exit(1);
			}
		}
	}

	/*
	cekani na dokonceni cinnosti vlaken/filosofu
	dokonceni probehne az kazdy filozof bude jist X krat (dle parametru)
	 */
	for(i = 0; i < pocet_vlaken; i++){
    pthread_join(vlakna[i], NULL);
	}

	/* uklid */
  for(i = 0; i < pocet_vlaken; i++){
    sem_destroy(&(vidlicky[i]));
	}
  free(vidlicky);
  free(jedl);

  return 0;
}

/*
Vlakno predstavujici filozofa, ktery premysli a veceri
TYP SIMULACE: standard
*/
void *filozof(void *argument){
	int cislo = (int) argument; /*Cislo aktualniho vlakna*/
	int filozof = cislo+1; /*Cislo filozofa*/
	int leva = cislo; /*Cislo leve vydlicky*/
	int prava = (cislo+1)%pocet_vlaken; /*Cislo prave vydlicky*/

	printf("Ahoj jsem filozof %02d\n", filozof);

	while(1){
		printf("%2d:\t Premyslim... \n", filozof);
		slofik();
		if(sem_trywait(&vidlicky_pristup) == 0){
	    printf("%2d:\t Muzu brat vidlicky\n", filozof);
			if(sem_trywait(&vidlicky[leva]) != 0){
				printf("%2d:\t Nemuzu vzit levou vidlicku\n", filozof);
				sem_post(&vidlicky_pristup);
			}
			else { /* mam levou vidlicku */
				if(sem_wait(&vidlicky[prava]) != 0){
					printf("%2d:\t Nemuzu vzit pravou vidlicku\n", filozof);
					sem_post(&vidlicky[leva]);
					sem_post(&vidlicky_pristup);
				}
				else { /* mam pravou vidlicky => muzu jist*/
		      printf("%2d:\t Mam vidlicky.\n", filozof);
					sem_post(&vidlicky_pristup);
					printf("%2d:\t Jim... \n", filozof);
					jedl[cislo]++;
					slofik(); /*uspani v mikro sekundach*/

					/* polozeni vydlicek */
					sem_post(&vidlicky[leva]);
				  printf("%2d:\t Uvolnuji levou vidlicku\n", filozof);
					sem_post(&vidlicky[prava]);
				  printf("%2d:\t Uvolnuji pravou vidlicku\n", filozof);

					printf("%2d:\t Dojedl jsem, jdu premyslet. \n", filozof);
				}
			}
		}else {
			printf("%2d:\t Nepodarilo se ziskat pristup k vidlickam.\n", filozof);
			slofik();
		}

		if(jedl[cislo] == pocet_jezeni){
			printf("%2d:\t Uz jsem jedl %dkrat!\n", filozof, jedl[cislo]);
			sleep_sec(1);
	    pthread_exit(0); /* ukonceni vlakna */
		}

	}
}

/*
Vlakno predstavujici filozofa, ktery premysli a veceri
TYP SIMULACE: deadlock
*/
void *deadlock(void *argument){
	int cislo = (int) argument; /*Cislo aktualniho vlakna*/
	int filozof = cislo+1; /*Cislo filozofa*/
	int leva = cislo; /*Cislo leve vydlicky*/
	int prava = (cislo+1)%pocet_vlaken; /*Cislo prave vydlicky*/

	printf("Ahoj jsem filozof %02d\n", filozof);/**/

	while(1){
		printf("%2d:\t Premyslim... \n", filozof);
		usleep(100000);

    printf("%2d:\t Beru levou vidlicku\n", filozof);
		sem_wait(&vidlicky[leva]);
    printf("%2d:\t Vzal jsem levou vidlicku\n", filozof);
		usleep(100000);

    printf("%2d:\t Beru pravou vidlicku\n", filozof);
		sem_wait(&vidlicky[prava]);
    printf("%2d:\t Vzal jse pravou vidlicku\n", filozof);
		usleep(100000);

	  printf("%2d:\t Mam vidlicky.\n", filozof);
		printf("%2d:\t Jim... \n", filozof);
		jedl[cislo]++;

		/* polozeni vydlicek */
		sem_post(&vidlicky[leva]);
	  printf("%2d:\t Uvolnuji levou vidlicku\n", filozof);
		sem_post(&vidlicky[prava]);
	  printf("%2d:\t Uvolnuji pravou vidlicku\n", filozof);

		printf("%2d:\t Dojedl jsem, jdu premyslet. \n", filozof);

		if(jedl[cislo] == pocet_jezeni){
			printf("%2d:\t Uz jsem jedl %dkrat!\n", filozof, jedl[cislo]);
			sleep_sec(1);
	    pthread_exit(0); /* ukonceni vlakna */
		}
	}
}

/*
Vlakno predstavujici filozofa, ktery premysli a veceri
TYP SIMULACE: vyhladoveni (starvation)
*/
void *vyhladoveni(void *argument){
	int cislo = (int) argument; /*Cislo aktualniho vlakna*/
	int filozof = cislo+1; /*Cislo filozofa*/
	int leva = cislo; /*Cislo leve vydlicky*/
	int prava = (cislo+1)%pocet_vlaken; /*Cislo prave vydlicky*/

	printf("Ahoj jsem filozof %02d\n", filozof);/**/

	while(1){
		printf("%2d:\t Premyslim... \n", filozof);

    printf("%2d:\t Beru levou vidlicku\n", filozof);
		sem_wait(&vidlicky[leva]);
    printf("%2d:\t Vzal jsem levou vidlicku\n", filozof);

    printf("%2d:\t Beru pravou vidlicku\n", filozof);
		if(sem_trywait(&vidlicky[prava]) != 0){
			printf("%2d:\t Nemuzu vzit pravou vidlicku\n \tPokladam tedy levou a vyckam.\n", filozof);
			sem_post(&vidlicky[leva]);
		}else { /* mam obe vidlicky muzu se najist */
	    printf("%2d:\t Vzal jse pravou vidlicku\n", filozof);

		  printf("%2d:\t Mam obe vidlicky.\n", filozof);
      usleep(10000);
			printf("%2d:\t Jim... \n", filozof);
			jedl[cislo]++;

			/* polozeni vydlicek */
			sem_post(&vidlicky[leva]);
	    printf("%2d:\t Uvolnuji levou vidlicku\n", filozof);
			sem_post(&vidlicky[prava]);
	    printf("%2d:\t Uvolnuji pravou vidlicku\n", filozof);

			printf("%2d:\t Dojedl jsem, jdu premyslet. \n", filozof);

			if(jedl[cislo] == pocet_jezeni){
				printf("%2d:\t Uz jsem jedl %dkrat!\n", filozof, jedl[cislo]);
				sleep_sec(1);
		    pthread_exit(0); /* ukonceni vlakna */
			}
		}
	}
}

/*
Vlakno predstavujici filozofa, ktery premysli a veceri
TYP SIMULACE: podle poradi
*/
void *poradi(void *argument){
	int cislo = (int) argument; /*Cislo aktualniho vlakna*/
	int filozof = cislo+1; /*Cislo filozofa*/
	int leva = cislo; /*Cislo leve vydlicky*/
	int prava = (cislo+1)%pocet_vlaken; /*Cislo prave vydlicky*/

	printf("Ahoj jsem filozof %02d\n", filozof);/**/

	while(1){
		printf("%2d:\t Premyslim... \n", filozof);
		slofik();

		if(filozof != pocet_vlaken){
	    printf("%2d:\t Beru levou vidlicku\n", filozof);
			sem_wait(&vidlicky[leva]);
	    printf("%2d:\t Vzal jsem levou vidlicku\n", filozof);
			slofik();

	    printf("%2d:\t Beru pravou vidlicku\n", filozof);
			sem_wait(&vidlicky[prava]);
	    printf("%2d:\t Vzal jse pravou vidlicku\n", filozof);
			slofik();
		}
		else {
	    printf("%2d:\t Beru pravou vidlicku\n", filozof);
			sem_wait(&vidlicky[prava]);
	    printf("%2d:\t Vzal jse pravou vidlicku\n", filozof);
			slofik();

			printf("%2d:\t Beru levou vidlicku\n", filozof);
			sem_wait(&vidlicky[leva]);
	    printf("%2d:\t Vzal jsem levou vidlicku\n", filozof);
			slofik();
		}

	  printf("%2d:\t Mam vidlicky.\n", filozof);
		printf("%2d:\t Jim... \n", filozof);
		jedl[cislo]++;

		/* polozeni vydlicek */
		if(filozof != pocet_vlaken){
      sem_post(&vidlicky[prava]);
	    printf("%2d:\t Uvolnuji pravou vidlicku\n", filozof);
			sem_post(&vidlicky[leva]);
	    printf("%2d:\t Uvolnuji levou vidlicku\n", filozof);
			
		}
		else {
    	sem_post(&vidlicky[leva]);
	    printf("%2d:\t Uvolnuji pravou vidlicku\n", filozof);
			sem_post(&vidlicky[prava]);
	    printf("%2d:\t Uvolnuji levou vidlicku\n", filozof);
		}
		printf("%2d:\t Dojedl jsem, jdu premyslet. \n", filozof);

		if(jedl[cislo] == pocet_jezeni){
			printf("%2d:\t Uz jsem jedl %dkrat!\n", filozof, jedl[cislo]);
			sleep_sec(1);
	    pthread_exit(0); /* ukonceni vlakna */
		}
	}
}

void help(){
	printf("Napoveda:\n");
	printf("Spusteni: phil.c <pocet filozofu> <pocet jedeni> [<typ simulace>]\n\n");
	printf("Priklad: ./phil.c 5 2 nebo ./phil.c 30 30 deadlock\n\n\n");
	printf("Pocet filozofu: 5 - 30 (implicitne 5)\n");
	printf("Pocet jedeni: 2 - 30 (implicitně 2)\n");
	printf("Typ simulace: bez parametru || deadlock || starvation || order\n");
}
